import wx
import math
from asc_viewer.viewport import Viewport
from asc_viewer.symbol_instance import SymbolInstance
import glob, os
import rtreelib as rt
from asc_viewer.symbol import Symbol, window_types
from asc_viewer.bounded_canvas import BoundedCanvas

font_size_factors = [0.625, 1, 1.5, 2, 2.5, 3.5, 5, 7]


class Net:
    """A net as used in LtSpice."""

    def __init__(self, name):
        self.name = name
        self.connections = []  # a list of (SymbolInstance, InstancePin, PinName) tuples
        self.type = None  # denotes spur type as string
        self.wires = set()


class Connection:
    """A connection between a net and an instance."""

    def __init__(self, instance, pin, pin_name):
        self.instance = instance
        self.pin = pin
        self.pin_name = pin_name


class WirePoint:
    """An endpoint of a wire."""

    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.direction = (
            None  # direction for orientation of symbols attached to endpoints
        )
        self.wires = []
        self.net = None  # is set after DFS connects wires to nets


class Wire:
    """A wire as drawn in LtSpice."""

    def __init__(self, x0, y0, x1, y1):
        self.x0 = x0
        self.y0 = y0
        self.x1 = x1
        self.y1 = y1
        self.net = None


class AscCanvas(BoundedCanvas, Viewport):
    """Displays an LtSpice schematic.

    Arguments:
    parent -- the parent window
    symbol_paths -- a list of path names where symbols are stored
    instance_name -- the instance name of this schematic, this is only useful it the schematic is an instantiated subcircuit
    """

    def __init__(self, parent, symbol_paths, instance_name=""):
        super().__init__(parent)

        self.instance_name = instance_name
        self.symbols = {}
        self.filename = None
        self.load_symbols(symbol_paths)

        # create some graphics primitives for later use
        self.fonts = []
        self.dc = wx.ClientDC(self)
        self.gc = wx.GraphicsContext.Create(self.dc)
        self.black_pen = wx.Pen(wx.Colour(0, 0, 0), width=2, style=wx.PENSTYLE_SOLID)
        self.red_pen = wx.Pen(wx.Colour(255, 0, 0), width=2, style=wx.PENSTYLE_SOLID)
        for i in range(4):
            font = self.create_font(font_size_factors[i])
            self.fonts.append(font)
        font_size = 0.8
        self.black_font = self.create_font(font_size, wx.BLACK)
        self.blue_font = self.create_font(font_size, wx.BLUE)
        self.red_font = self.create_font(font_size, wx.RED)
        self.gray_font = self.create_font(font_size, wx.Colour(50, 50, 50, 50))

        self.wires = []
        self.wire_points = {}
        self.net_counter = 0  # for auto-labeling nets
        self.flags = {}  # off-schematic connectors or io pins
        self.texts = []
        self.rtree = rt.RTree()
        self.wire_lookup = rt.RTree()
        self.nets = {}  # name to net
        self.path = self.gc.CreatePath()
        self.symbol_instances = {}

        self.find_data = wx.FindReplaceData()
        self.find_dialog = None  # cannot be initialized here yet

        self.Bind(wx.EVT_CHAR_HOOK, self.on_key)
        self.Bind(wx.EVT_PAINT, self.on_paint)

    def load_symbols(self, symbol_paths):
        """Loads symbols from a list of paths to asy files."""
        if isinstance(symbol_paths, str):
            symbol_paths = [symbol_paths]
        for path in symbol_paths:
            path = os.path.join(path, "*.asy")
            filenames = glob.glob(path)
            for f in filenames:
                name = os.path.basename(f)[:-4]
                self.symbols[name] = Symbol(self, f)

    def create_font(self, size, color=wx.BLACK):
        return self.gc.CreateFont(
            wx.Font(
                int(10 * size),
                wx.FONTFAMILY_DEFAULT,
                wx.FONTSTYLE_NORMAL,
                wx.FONTWEIGHT_NORMAL,
            ),
            col=color,
        )

    def align_text(self, x, y, text, align, size, rotation, morig):
        """Aligns text for presentation. It takes formatting inputs from the asc file,
        i.e., rotation, alignment and matrix transformation and returns coordinates suitable
        for wx's DrawText function."""
        m = self.gc.CreateMatrix()
        m.Concat(morig)
        self.gc.SetFont(self.fonts[size])
        w, h, d, e = self.gc.GetFullTextExtent(text)
        if align[0] == "V":
            align = align[1:]
            m.Translate(x, y)
            m.Rotate(math.pi / 2 * 3)
            m.Translate(-x, -y)
            rotation += 270
        if align == "Right":
            x -= w
            y -= h / 2
        elif align == "Left":
            y -= h / 2
        elif align == "Top":
            x -= w / 2
        elif align == "Bottom":
            x -= w / 2
            y -= h
        elif align == "Center":
            x -= w / 2
            y -= h / 2
        m.Translate(x, y)

        # the only valid text directions are right or up
        if rotation % 360 == 180 or rotation % 360 == 90:
            m.Translate(w / 2, h / 2)
            m.Rotate(math.pi)
            m.Translate(-w / 2, -h / 2)
        x1, y1 = m.TransformPoint(0, 0)
        x2, y2 = m.TransformPoint(w, h)
        return min(x1, x2), min(y1, y2), max(y1, y2)

    def connect_wires(self, wire_point):
        """Connects wires to nets using recursion."""
        stack = [wire_point]
        while stack:
            wire_point = stack.pop()
            # check for a custom net name
            flag = self.flags.get((wire_point.x, wire_point.y))
            if flag:
                assert (
                    self.net_override is None or self.net_override == flag["net"]
                ), f"Conflicting net names are assigned: {self.net_override} and {flag['net']}"
                if self.net_override is None:
                    self.net_counter -= 1
                self.net_override = flag["net"]
                wire_point.net.name = flag["net"]

            for wire in wire_point.wires:
                wire.net = wire_point.net
                wire_point.net.wires.add(wire)
                if wire_point.x == wire.x0 and wire_point.y == wire.y0:
                    neighbor = self.wire_points.get((wire.x1, wire.y1))
                else:
                    neighbor = self.wire_points.get((wire.x0, wire.y0))
                if neighbor.net is None:
                    neighbor.net = wire_point.net
                    stack.append(neighbor)

    def load_asc(self, filename):
        """Loads an LtSpice schematic from the given filename."""
        instances = []
        self.filename = filename
        f = open(filename, encoding="iso-8859-1")
        self.reset_extent()
        sheet_w, sheet_h = 0, 0
        for line in f:
            line = line.strip()
            if len(line) == 0:
                continue
            words = line.split(" ")
            if words[0] == "WIRE":
                wire = Wire(*[int(x) for x in words[1:]])
                self.check_extent([wire.x0, wire.x1], [wire.y0, wire.y1])
                self.wires.append(wire)

                # save endpoints and determine direction of wire ends that is used for some connector symbols and ground
                # default direction at wire end is down
                wire_point0 = self.wire_points.get((wire.x0, wire.y0))
                if wire_point0 is None:
                    wire_point0 = WirePoint(wire.x0, wire.y0)
                    self.wire_points[(wire.x0, wire.y0)] = wire_point0
                wire_point1 = self.wire_points.get((wire.x1, wire.y1))
                if wire_point1 is None:
                    wire_point1 = WirePoint(wire.x1, wire.y1)
                    self.wire_points[(wire.x1, wire.y1)] = wire_point1
                wire_point0.wires.append(wire)
                wire_point1.wires.append(wire)
                if wire.x0 == wire.x1:  # vertical
                    if wire.y0 < wire.y1:
                        wire_point0.direction = 2  # top
                    else:
                        wire_point1.direction = 2
                if wire.y0 == wire.y1:  # horizontal
                    if wire.x0 < wire.x1:
                        wire_point0.direction = 1  # left
                        wire_point1.direction = 3
                    else:
                        wire_point0.direction = 3  # right
                        wire_point1.direction = 1

                min_x, min_y = min(wire.x0, wire.x1), min(
                    wire.y0, wire.y1
                )  # rtree lib needs a well-formed rect
                max_x, max_y = max(wire.x0, wire.x1), max(wire.y0, wire.y1)
                rect = rt.Rect(min_x, min_y, max_x + 1, max_y + 1)
                self.wire_lookup.insert(wire, rect)
            elif words[0] == "TEXT":
                x = int(words[1])
                y = int(words[2])
                align = words[3]
                size = int(words[4])
                text = " ".join(words[5:])
                self.check_extent(x - 15, y - 15)

                x, y, y2 = self.align_text(
                    x, y, text, align, size, 0, self.gc.CreateMatrix()
                )
                t = dict(x=x, y=y, size=size, text=text)

                self.texts.append(t)
            elif words[0] == "SHEET":
                sheet_w, sheet_h = int(words[2]), int(words[3])
            elif words[0] == "FLAG":
                last_flag = dict(
                    x=int(words[1]), y=int(words[2]), net=words[3], type=None
                )
                self.check_extent(last_flag["x"], last_flag["y"])
                self.flags[(last_flag["x"], last_flag["y"])] = last_flag
            elif words[0] == "IOPIN":
                last_flag["type"] = words[3]
            elif words[0] == "SYMATTR":
                attr = " ".join(words[2:])
                if words[1] == "InstName" and self.instance_name != "":
                    attr = self.instance_name + "." + attr
                instances[-1].attrs[words[1]] = attr
            elif words[0] == "SYMBOL":
                instance = SymbolInstance(
                    self,
                    words[1],
                    int(words[2]),
                    int(words[3]),
                    words[4][0] == "M",
                    int(words[4][1:]),
                )
                instances.append(instance)
                self.check_extent(instance.x, instance.y)
                # load default attrs from symbol file
                symbol = self.symbols.get(words[1])
                assert symbol, f"Unknown symbol {words[1]}"
                symbol.load()
                instance.attrs = symbol.attrs.copy()
            elif words[0] == "WINDOW":
                x = int(words[2])
                y = int(words[3])
                window = dict(
                    type=window_types[words[1]],
                    x=x,
                    y=y,
                    align=words[4],
                    size=int(words[5]),
                )
                instances[-1].windows[window["type"]] = window

        # load symbol instances
        self.symbol_instances = {}
        pin_positions = {}
        for instance in instances:
            s = self.symbols.get(instance.name)
            if s is None:
                print(f"Symbol not found {instance.name}")
                self.rtree.insert(
                    instance,
                    rt.Rect(
                        instance.x - 5, instance.y - 5, instance.x + 5, instance.y + 5
                    ),
                )
                continue
            s.load()
            instance.set_symbol(s)
            for pin in instance.pins:
                pin_positions[(pin.x, pin.y)] = (instance, pin)
            x1, y1 = instance.matrix.TransformPoint(s.x1, s.y1)
            x2, y2 = instance.matrix.TransformPoint(s.x2, s.y2)
            if x1 > x2:
                x1, x2 = x2, x1
            if y1 > y2:
                y1, y2 = y2, y1
            self.rtree.insert(
                instance,
                rt.Rect(
                    instance.x + x1, instance.y + y1, instance.x + x2, instance.y + y2
                ),
            )
            self.symbol_instances[instance.attrs["InstName"]] = instance
            self.check_extent(
                [instance.x + s.x1, instance.x + s.x2],
                [instance.y + s.y1, instance.y + s.y2],
            )

        # connect wires to pins
        for wire_point in self.wire_points.values():
            if wire_point.net:
                continue
            self.net_counter += 1
            wire_point.net = Net(f"N{self.net_counter:03d}")
            self.nets[wire_point.net.name] = wire_point.net
            self.net_override = (
                None  # is set to net name if a user-assigned net name is found
            )
            self.connect_wires(wire_point)

        # calculate correct pin index from SpiceOrder

        # add instance connections to nets and label spur types
        for wire_point in self.wire_points.values():
            res = pin_positions.get((wire_point.x, wire_point.y))
            if res:
                instance, pin = res
                pin_name = str(pin.symbol_pin.index)
                connection = Connection(instance, pin, pin_name)
                wire_point.net.connections.append(connection)

        # add net flags to main rtree
        for flag in self.flags.values():
            x1 = flag["x"]
            y1 = flag["y"]
            x2 = x1 + 20
            y2 = y1 + 20
            net = self.nets.setdefault(flag["net"], Net(flag["net"]))
            self.rtree.insert(net, rt.Rect(x1, y1, x2, y2))

        self.x1 -= 10
        self.y1 -= 10
        self.x2 = max(self.x2, sheet_w)
        self.y2 = max(self.y2, sheet_h)
        self.set_size(self.x2 - self.x1, self.y2 - self.y1)
        self.Refresh()

        self.path = self.gc.CreatePath()

        for wire in self.wires:
            self.path.MoveToPoint(wire.x0, wire.y0)
            self.path.AddLineToPoint(wire.x1, wire.y1)

        for wire_point in self.wire_points.values():
            # add dots representing wire connection
            if len(wire_point.wires) > 2:
                self.path.AddRectangle(wire_point.x - 2, wire_point.y - 2, 4, 4)

        for flag in self.flags.values():
            x1, y1 = flag["x"], flag["y"]
            if flag["type"] == "In":
                path = self.gc.CreatePath()
                path.MoveToPoint(x1, y1)
                path.AddLineToPoint(x1 + 10, y1 + 10)
                path.AddLineToPoint(x1 + 10, y1 + 20)
                path.AddLineToPoint(x1 - 10, y1 + 20)
                path.AddLineToPoint(x1 - 10, y1 + 10)
                path.AddLineToPoint(x1, y1)
                wire_point = self.wire_points.get((x1, y1))
                if wire_point and len(wire_point.wires) == 1:
                    direction = wire_point.direction
                    if direction:
                        m = self.gc.CreateMatrix()
                        m.Translate(x1, y1)
                        m.Rotate(math.pi / 2 * direction)
                        m.Translate(-x1, -y1)
                        path.Transform(m)
                self.path.AddPath(path)
            elif flag["type"] == "Out":
                pass
            elif flag["type"] == "BiDir":
                pass
            elif flag["net"] == "0":
                path = self.gc.CreatePath()
                path.MoveToPoint(x1 - 10, y1)
                path.AddLineToPoint(x1 + 10, y1)
                path.MoveToPoint(x1 - 10, y1)
                path.AddLineToPoint(x1, y1 + 10)
                path.MoveToPoint(x1 + 10, y1)
                path.AddLineToPoint(x1, y1 + 10)
                self.path.AddPath(path)

    def mouse_position(self, evt):
        """Returns the mouse position in schematic canvas coordinates."""
        dc = wx.ClientDC(self)
        self.DoPrepareDC(dc)
        x, y = evt.GetLogicalPosition(dc)
        return (x + self.x1, y + self.y1)

    def get_instance_under_mouse(self, evt):
        """Returns the symbol instance under the mouse pointer."""
        if len(self.symbol_instances) == 0:
            return None
        pos = self.mouse_position(evt)
        res = self.rtree.query(pos)
        try:
            return next(res).data
        except StopIteration:
            return None

    def get_net_under_mouse(self, evt):
        """Returns the net under the mouse pointer."""
        pos = self.mouse_position(evt)
        rect = (pos[0] - 5, pos[1] - 5, pos[0] + 5, pos[1] + 5)
        res = self.wire_lookup.query(rect)

        def distance_to_line(x0, y0, x1, y1, x2, y2):
            return (
                abs((x2 - x1) * (y1 - y0) - (x1 - x0) * (y2 - y1))
                / ((x2 - x1) ** 2 + (y2 - y1) ** 2) ** 0.5
            )

        res = [x.data for x in res]
        if not res:
            return None
        closest_wire = min(
            res,
            key=lambda wire: distance_to_line(
                pos[0], pos[1], wire.x0, wire.y0, wire.x1, wire.y1
            ),
        )
        return closest_wire.net

    def on_key(self, evt):
        k = evt.GetUnicodeKey()
        if evt.ControlDown():
            k = chr(k)
            if k == "F":
                if self.find_dialog is None:
                    title = f"Find in {self.instance_name}"
                    self.find_data.SetFindString("")
                    self.find_dialog = wx.FindReplaceDialog(None, self.find_data, title)
                    self.find_dialog.Bind(wx.EVT_FIND, self.on_find)
                self.find_dialog.Show()
                return
        evt.Skip()

    def on_find(self, evt):
        find = evt.GetFindString()
        self.find_dialog.Destroy()
        self.find_dialog = None

        # search nodes
        r = list(self.rtree.search(None, lambda x: x.data.name == find))
        if len(r):
            r = r[0]
            self.center_on(r.rect.min_x - self.x1, r.rect.min_y - self.y1)
        else:
            wx.MessageDialog(
                None, "No matches", "Error", wx.OK | wx.ICON_QUESTION
            ).ShowModal()

    def go_to_edge(self, instance_name):
        """Scrolls the canvas to the symbol instance specified by instance_name."""
        if not instance_name.startswith(self.instance_name):
            print(f"{instance_name} is not in {self.instance_name}")
            return
        local_instance = instance_name[len(self.instance_name) + 1 :].split(".")[0]
        instance_name = self.instance_name + "." + local_instance
        s = self.symbol_instances.get(instance_name)
        if s is None:
            return
        self.center_on(s.x - self.x1, s.y - self.y1)

    def on_paint(self, evt):
        """Paints the schematic."""
        dc = wx.PaintDC(self)
        self.DoPrepareDC(dc)
        gc = wx.GraphicsContext.Create(dc)
        gc.Translate(-self.x1, -self.y1)

        gc.SetPen(self.black_pen)
        gc.StrokePath(self.path)
        gc.SetFont(self.fonts[1])
        for flag in self.flags.values():
            if flag["net"] == "0":
                continue
            gc.DrawText(flag["net"], flag["x"], flag["y"])

        for instance in self.symbol_instances.values():
            instance.paint(gc)

        for text in self.texts:
            gc.SetFont(self.fonts[text["size"]])
            gc.DrawText(text["text"], text["x"], text["y"])



import wx


class Viewport(wx.ScrolledCanvas):
    """An abstract viewport that supports scrolling and zooming, without implementing painting."""

    def __init__(self, *args):
        wx.ScrolledCanvas.__init__(self, *args)
        self.zoom = 1.0

        self.Bind(wx.EVT_MOUSEWHEEL, self.OnWheel)
        self.Bind(wx.EVT_MIDDLE_DOWN, self.OnMiddleDown)
        self.Bind(wx.EVT_MIDDLE_UP, self.OnMiddleUp)
        self.Bind(wx.EVT_MOTION, self.OnMotion)
        self.dragging = ""
        self.w = 1
        self.h = 1
        self.bmp = None
        self.SetScrollRate(1, 1)

    def set_size(self, width, height):
        self.w = width
        self.h = height
        self.zoom = 1
        # zoom to fit:
        #  w, h = self.GetClientSize()
        #  self.zoom = min(w/self.w, h/self.h)
        self.SetScale(self.zoom, self.zoom)
        self.SetVirtualSize(int(self.w * self.zoom), int(self.h * self.zoom))

    def get_scroll_origin(self):
        return [x / self.zoom for x in self.GetViewStart()]

    def set_scroll_origin(self, x, y):
        self.Scroll(int(x * self.zoom), int(y * self.zoom))

    def center_on(self, x, y):
        w, h = self.GetClientSize()
        w /= 2 * self.zoom
        h /= 2 * self.zoom
        x -= w
        y -= h
        self.Scroll(int(x * self.zoom), int(y * self.zoom))

    def set_zoom(self, zoom):
        """Changes scroll origin."""
        if zoom == "full":
            w, h = self.GetClientSize()
            zoom = min(w / self.w, h / self.h)
        self.SetScale(zoom, zoom)
        self.SetVirtualSize(int(self.w * zoom), int(self.h * zoom))
        self.zoom = zoom
        self.Refresh()

    def OnMiddleDown(self, evt):
        dc = wx.ClientDC(self)
        self.DoPrepareDC(dc)
        pos = evt.GetLogicalPosition(dc)
        self.dragging = "middle"
        self.scroll_origin_x, self.scroll_origin_y = self.GetViewStart()
        self.scroll_origin_x /= self.zoom
        self.scroll_origin_y /= self.zoom
        self.drag_origin = pos
        self.drag_pos = pos

    def OnMiddleUp(self, evt):
        self.dragging = ""

    def OnMotion(self, evt):
        if not evt.Dragging():
            return
        dc = wx.ClientDC(self)
        self.DoPrepareDC(dc)
        pos = evt.GetLogicalPosition(dc)
        if self.dragging == "middle":
            x = self.scroll_origin_x + (self.drag_origin[0] - pos[0])
            y = self.scroll_origin_y + (self.drag_origin[1] - pos[1])
            if x < 0:
                self.drag_origin[0] -= int(x)
            if y < 0:
                self.drag_origin[1] -= int(y)
            w, h = self.GetClientSize()
            w /= self.zoom
            h /= self.zoom
            if x > self.w - w:
                self.drag_origin[0] += int(self.w - w - x)
            if y > self.h - h:
                self.drag_origin[1] += int(self.h - h - y)
            self.scroll_origin_x = max(min(x, self.w - w), 0)
            self.scroll_origin_y = max(min(y, self.h - h), 0)
            self.Scroll(
                int(self.scroll_origin_x * self.zoom),
                int(self.scroll_origin_y * self.zoom),
            )
        self.Refresh()
        self.drag_pos = pos

    def OnWheel(self, evt):
        dc = wx.ClientDC(self)
        self.DoPrepareDC(dc)
        pos = evt.GetLogicalPosition(dc)
        offset = self.CalcScrolledPosition(pos[0], pos[1])

        zoom = self.zoom * 1.2 ** (evt.GetWheelRotation() / 120)
        if (
            zoom < 0.2
            and evt.GetWheelRotation() < 0
            or zoom > 16
            and evt.GetWheelRotation() > 0
        ):
            return
        scroll_origin_x, scroll_origin_y = self.GetViewStart()
        scroll_origin_x /= self.zoom
        scroll_origin_y /= self.zoom
        self.SetScale(zoom, zoom)
        self.SetVirtualSize(int(self.w * zoom), int(self.h * zoom))
        scroll_origin_x = int(self.zoom * (scroll_origin_x - pos[0]) + pos[0] * zoom)
        scroll_origin_y = int(self.zoom * (scroll_origin_y - pos[1]) + pos[1] * zoom)
        self.Scroll(scroll_origin_x, scroll_origin_y)
        self.zoom = zoom
        self.Refresh()



from asc_viewer.asc_canvas import AscCanvas, Net, Connection
from asc_viewer.symbol import Symbol
from asc_viewer.symbol_instance import SymbolInstance, Pin
from asc_viewer.bounded_canvas import BoundedCanvas
from asc_viewer.viewport import Viewport



import wx
import math
from asc_viewer.bounded_canvas import BoundedCanvas

window_types = {
    "0": "InstName",
    "3": "Value",
    "39": "SpiceLine",
    "123": "Value2",
    "?": "SpiceLine2",
}


class Pin:
    """An LtSpice symbol pin as defined in an .asy file"""

    def __init__(self):
        self.name = None
        self.order = None
        self.align = None
        self.x = None
        self.y = None
        self.text_x = None
        self.text_y = None

    def rect(self):
        return (self.x - 5, self.y - 5, 10, 10)


class Symbol(BoundedCanvas):
    """An LtSpice symbol as defined in an .asy file"""

    def __init__(self, parent, filename):
        """Arguments:
        parent -- A wx class that provides a graphics context, i.e., AscViewer
        filename -- the full filename of the .asy file
        """
        super().__init__()
        self.filename = filename
        self.loaded = False
        self.lines = []
        self.circles = []
        self.arcs = []
        self.windows = {}
        self.pins = []
        self.circles = []
        self.texts = []
        self.rectangles = []
        self.attrs = {}
        self.parent = parent

    def load(self):
        """Loads the symbol from file."""
        if self.loaded:
            return
        dc = wx.ClientDC(self.parent)
        gc = wx.GraphicsContext.Create(dc)

        f = open(self.filename, encoding="iso-8859-1")
        self.reset_extent()
        last_pin = None
        for line in f:
            line = line.strip()
            if len(line) == 0:
                continue
            words = line.split(" ")
            if words[0] == "SymbolType":
                self.type = words[1]
            elif words[0] == "LINE":
                line = dict(style=words[1], coords=[int(x) for x in words[2:]])
                self.check_extent(
                    [line["coords"][0], line["coords"][2]],
                    [line["coords"][1], line["coords"][3]],
                )
                self.lines.append(line)
            elif words[0] == "CIRCLE":
                c = [int(x) for x in words[2:]]
                self.check_extent([c[0], c[2]], [c[1], c[3]])
                c = (
                    min(c[0], c[2]),
                    min(c[1], c[3]),
                    abs(c[2] - c[0]),
                    abs(c[3] - c[1]),
                )
                c = dict(style=words[1], coords=c)
                self.circles.append(c)
            elif words[0] == "ARC":
                c = [int(x) for x in words[2:]]
                self.check_extent([c[0], c[2]], [c[1], c[3]])
                x1, y1, w, h = (
                    min(c[0], c[2]),
                    min(c[1], c[3]),
                    abs(c[2] - c[0]),
                    abs(c[3] - c[1]),
                )
                cx, cy, rx, ry = (x1 + w / 2, y1 + h / 2, w / 2, h / 2)
                angle1 = int(math.atan2(c[5] - cy, c[4] - cx) * 180 / math.pi)
                angle2 = int(math.atan2(c[7] - cy, c[6] - cx) * 180 / math.pi)
                if angle2 > angle1:
                    angle2 -= 360
                coords = [cx, cy, rx, ry, angle1, angle2]
                c = dict(style=words[1], coords=coords)
                self.arcs.append(c)
            elif words[0] == "TEXT":
                x = int(words[1])
                y = int(words[2])
                align = words[3]
                size = int(words[4])
                text = " ".join(words[5:])
                self.check_extent(x - 15, y - 15)

                x, y, y2 = self.parent.align_text(
                    x, y, text, align, size, 0, gc.CreateMatrix()
                )
                t = dict(x=x, y=y, size=size, text=text)
                self.texts.append(t)
            elif words[0] == "WINDOW":
                x = int(words[2])
                y = int(words[3])
                type = window_types[words[1]]
                window = dict(type=type, x=x, y=y, align=words[4], size=int(words[5]))
                self.check_extent(x - 15, y - 15)
                self.windows[type] = window
            elif words[0] == "RECTANGLE":
                c = [int(x) for x in words[2:]]
                self.check_extent([c[0], c[2]], [c[1], c[3]])
                c = [
                    min(c[0], c[2]),
                    min(c[1], c[3]),
                    abs(c[2] - c[0]),
                    abs(c[3] - c[1]),
                ]
                rect = dict(style=words[1], coords=c)
                self.rectangles.append(rect)
            elif words[0] == "PIN":
                last_pin = Pin()
                last_pin.x = int(words[1])
                last_pin.y = int(words[2])
                map_align = {
                    "left": "Right",
                    "right": "Left",
                    "top": "Bottom",
                    "bottom": "Top",
                    "none": "",
                }
                last_pin.align = words[3].capitalize()
                offset = int(words[4])
                last_pin.text_x = last_pin.x
                last_pin.text_y = last_pin.y
                if last_pin.align == "Top":
                    last_pin.text_y = last_pin.y + offset
                elif last_pin.align == "Bottom":
                    last_pin.text_y = last_pin.y - offset
                elif last_pin.align == "Left":
                    last_pin.text_x = last_pin.x + offset
                elif last_pin.align == "Right":
                    last_pin.text_x = last_pin.x - offset
                self.check_extent(last_pin.x, last_pin.y)
                self.pins.append(last_pin)
                # self.rectangles.append(dict(style='Normal', coords=c))
            elif words[0] == "PINATTR":
                assert last_pin
                key, value = words[1], words[2]
                if key == "PinName":
                    last_pin.name = value
                elif key == "SpiceOrder":
                    last_pin.order = int(value)
                else:
                    assert f"Unknown pin attribute {key}"
            elif words[0] == "SYMATTR":
                self.attrs[words[1]] = " ".join(words[2:])
        self.loaded = True

        # assign zero-based pin indices
        self.pins.sort(key=lambda pin: pin.order)
        for i, pin in enumerate(self.pins):
            pin.index = i

        # create the path that draws the symbol
        path = self.parent.gc.CreatePath()
        for line in self.lines:
            c = line["coords"]
            path.MoveToPoint(c[0], c[1])
            path.AddLineToPoint(c[2], c[3])
        for circle in self.circles:
            path.AddEllipse(*circle["coords"])
        for arc in self.arcs:
            c = arc["coords"]
            path.MoveToPoint(
                c[0] + c[2] * math.cos(c[4] / 180 * math.pi),
                c[1] + c[3] * math.sin(c[4] / 180 * math.pi),
            )
            for a in range(c[4], c[5], -10):
                path.AddLineToPoint(
                    c[0] + c[2] * math.cos(a / 180 * math.pi),
                    c[1] + c[3] * math.sin(a / 180 * math.pi),
                )
        self.path = path

    def paint(self, gc, old_m, rotation, attrs, windows):
        """Paints the symbol in its instantiated representation as part of a larger schematic.
        Arguments:

        gc -- A wx graphics context.
        old_m -- A matrix to be used for drawing the symbol, rotation is supplied separately.
        rotation -- Symbol rotation in degrees.
        attrs -- Attributes of the symbol instance.
        windows -- Windows supplied by the symbold instance.
        """
        gc.StrokePath(self.path)
        for text in self.texts:
            gc.SetFont(self.parent.fonts[text["size"]])
            gc.DrawText(text["text"], text["x"], text["y"])
        for rect in self.rectangles:
            gc.DrawRectangle(*rect["coords"])

        combined_windows = self.windows | windows

        m = gc.GetTransform()
        gc.SetTransform(old_m)  # un-rotated transform
        # get text rotation matrix
        text_m = gc.CreateMatrix(*old_m.Get())
        text_m.Invert()
        text_m.Concat(m)

        for pin in self.pins:
            if not pin.name:
                continue  # unnamed pin
            if not pin.align or pin.align == "None":
                continue  # hidden pin
            x, y, y2 = self.parent.align_text(
                pin.text_x, pin.text_y, pin.name, pin.align, 2, 0, text_m
            )  # pin names have a fixed size of 1.5
            gc.DrawText(pin.name, x, y, angle=0)

        for type, window in combined_windows.items():
            gc.SetFont(self.parent.fonts[window["size"]])
            text = attrs.get(type)
            if text is None:
                text = self.attrs.get(type, "NA")  # use default attr from symbol
            x, y, y2 = self.parent.align_text(
                window["x"],
                window["y"],
                text,
                window["align"],
                window["size"],
                rotation,
                text_m,
            )
            angle = 90 if window["align"][0] == "V" else 0
            if rotation == 270 or rotation == 90:
                angle -= 90
            if angle:
                y = y2
            gc.DrawText(text, x, y, angle=angle / 180 * math.pi)



class BoundedCanvas:
    """A mixin that keeps track of canvas size."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reset_extent()

    def reset_extent(self):
        """Resets the size to 0."""
        self.x1 = 0
        self.x2 = 0
        self.y1 = 0
        self.y2 = 0
        self.w = 0
        self.h = 0

    def check_extent(self, xs, ys):
        """Grows the canvas if needed such that the given coordinates fit inside it.

        Arguments:
        xs --- list of x values to check.
        ys --- list of y values to check.
        """
        if isinstance(xs, int):
            xs = [xs]
        if isinstance(ys, int):
            ys = [ys]
        for x in xs:
            if x < self.x1:
                self.x1 = x
            if x > self.x2:
                self.x2 = x
        for y in ys:
            if y < self.y1:
                self.y1 = y
            if y > self.y2:
                self.y2 = y
        self.w = self.x2 - self.x1
        self.h = self.y2 - self.y1

    def get_extent(self):
        return self.x1, self.y1, self.x2, self.y2



import io
import os
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

setup(
    name="asc_viewer",
    version="1.0",
    description="A viewer for LTspice ASC schematics implemented in wxPython",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='http://github.com/ahaensler/asc_viewer',
    scripts=["bin/asc_viewer"],
    packages=["asc_viewer"],
    author="Adrian Haensler",
    license='MIT',
    python_requires=">=3.8.0",
    install_requires=[
        "rtreelib>=0.2.0",
        "wxPython>=4.2.0",
    ],
)



import wx
import math


class Pin:
    def __init__(self, symbol_pin):
        self.symbol_pin = symbol_pin
        self.x = None
        self.y = None


class SymbolInstance:
    def __init__(self, parent, name, x, y, mirror, rotation):
        self.prefix = parent.instance_name
        self.name = name
        self.x = x
        self.y = y
        self.rotation = rotation
        self.mirror = mirror
        self.attrs = {}
        self.symbol = None
        self.pins = []
        self.windows = {}

        self.matrix = parent.gc.CreateMatrix()
        if self.mirror:
            a, b, c, d, x, y = self.matrix.Get()
            self.matrix.Set(a=-a)
        self.matrix.Rotate(self.rotation / 180 * math.pi)

        self.user_data = None  # links arbitrary user data to this instance

        self.black_pen = wx.Pen(wx.Colour(0, 0, 0), width=2, style=wx.PENSTYLE_SOLID)
        self.red_pen = wx.Pen(wx.Colour(255, 0, 0), width=2, style=wx.PENSTYLE_SOLID)
        self.blue_pen = wx.Pen(wx.Colour(0, 255, 0), width=2, style=wx.PENSTYLE_SOLID)
        self.no_pen = wx.Pen(wx.Colour(0, 0, 0), style=wx.PENSTYLE_TRANSPARENT)
        self.orange_brush = wx.Brush(
            wx.Colour(250, 150, 150), style=wx.BRUSHSTYLE_SOLID
        )
        self.no_brush = wx.Brush(wx.Colour(0, 0, 0), style=wx.BRUSHSTYLE_TRANSPARENT)
        self.user_paint = (
            lambda self, gc, user_data: None
        )  # hook for a user-defined paint function

    def set_symbol(self, symbol):
        self.symbol = symbol
        for symbol_pin in self.symbol.pins:
            pin = Pin(symbol_pin)
            dx, dy = self.matrix.TransformPoint(symbol_pin.x, symbol_pin.y)
            pin.x = self.x + round(
                dx
            )  # rounding is important for pins and wires to line up and connect
            pin.y = self.y + round(dy)
            self.pins.append(pin)

    def get_extent(self):
        # returns the extent of the instance on the canvas
        x0, y0, x1, y1 = self.symbol.get_extent()
        x0, y0 = self.matrix.TransformPoint(x0, y0)
        x1, y1 = self.matrix.TransformPoint(x1, y1)
        x0 += self.x
        x1 += self.x
        y0 += self.y
        y1 += self.y
        # add space to accomodate pen width
        return min(x0, x1) - 2, min(y0, y1) - 2, max(x0, x1) + 2, max(y0, y1) + 2

    def set_user_data(self, user_data):
        self.user_data = user_data

    def set_user_paint_func(self, func):
        self.user_paint = func

    def paint(self, gc):
        old_m = gc.GetTransform()
        gc.Translate(self.x, self.y)
        old_m2 = gc.GetTransform()  # translated to symbol position, but no rotation

        # concat order was inconsistent (https://github.com/wxWidgets/wxWidgets/issues/17670), fixed now
        m2 = gc.CreateMatrix(*old_m2.Get())
        m2.Concat(self.matrix)
        gc.SetTransform(m2)

        if not self.symbol:
            return
        self.user_paint(self, gc, self.user_data)

        self.symbol.paint(gc, old_m2, self.rotation, self.attrs, self.windows)
        gc.SetBrush(self.no_brush)
        gc.SetPen(self.black_pen)
        gc.SetTransform(old_m)
